/**
 * Author Murukesh Sadasivan (murukesh@codeflow.co)
 *
 * Copyright 2016 Codeflow Private Limited
 *
 * MIT Licensed
 */
'use strict'

var Http = require('follow-redirects').http
var Https = require('follow-redirects').https
var Url = require('url')
var querystring = require("querystring")
var FormData = require('form-data')
var crypto = require("crypto")

module.exports.applyInputFilter = async (_input, callback, options) => {
    const input = _input || {}
    let uri = Url.parse(input.url || "", true, true)
    let inputSchema = { ...options.step['input-schema'] }
    let outputSchema = { ...options.step['output-schema'] }
    //set url description based on computed query params
    let url = "${url}"
    if (input.parameters && typeof input.parameters.forEach == 'function') {
        input.parameters.forEach((parameter, idx) => {
            if (parameter.key) {
                url += (idx > 0 || (uri.search && uri.search.length > 1) ? '&' : '?')
                url += parameter.key + "=" + encodeURIComponent(parameter.value ? parameter.value.toString() : '')
            }
        })
    }
    // inputSchema.properties.url.description = url

    //show body only for non-get
    const bodySchema = inputSchema.properties.body
    input.method = input.method || 'GET'
    if (input.method && input.method === 'GET') {
        bodySchema._hidden = true
    } else {
        delete bodySchema._hidden
        let showFormData = false
        let showContentType = false
        let showAttachments = false
        let showData = false
        let showFileUpload = false
        const body = input.body || {}
        if (body.type === "raw") {
            showContentType = true
            showData = true
        } else if (body.type === "file") {
            showFileUpload = true
        } else if (body.type == "form-data") {
            showAttachments = true
            showFormData = true
        } else if (body.type == "x-www-form-urlencoded") {
            showFormData = true
        }
        if (showFormData) {
            delete bodySchema.properties.formData._hidden
            if (showAttachments) {
                delete bodySchema.properties.formData.properties.files._hidden
            } else {
                bodySchema.properties.formData.properties.files._hidden = true
            }
        } else {
            bodySchema.properties.formData._hidden = true
        }
        if (showContentType) {
            delete bodySchema.properties.contentType._hidden
        } else {
            bodySchema.properties.contentType._hidden = true
        }
        if (showData) {
            delete bodySchema.properties.data._hidden
        } else {
            bodySchema.properties.data._hidden = true
        }
        if (showFileUpload) {
            delete bodySchema.properties.binaryData._hidden
        } else {
            bodySchema.properties.binaryData._hidden = true
        }
    }
    const advanced = input.advanced || {}
    let responseType = { type: "object" }
    if (advanced.parseAsJson === true) {
        delete inputSchema.properties.advanced.properties.outputSchema._hidden
        if (advanced.outputSchema) {
            responseType = { $ref: "file://" + advanced.outputSchema }
        } else {
            responseType = { type: "any" }
        }
    } else {
        inputSchema.properties.advanced.properties.outputSchema._hidden = true
        if (advanced.encoding && advanced.encoding !== 'binary') {
            responseType = { type: "string" }
        }
    }
    outputSchema.properties.responseData = responseType
    return callback(null, { inputSchema, outputSchema })
}


module.exports.doRestRequest = function(_input, callback) {
    const input = _input || {}
    if (!input.url || input.url === '') {
        return callback(new Error("url must be specified in the input."))
    }
    let uri = Url.parse(input.url, true, true),
        path = uri.path,
        headers = {}

    if (input.headers) {
        input.headers && input.headers.forEach(header => {
            if (header.key) {
                headers[header.key] = header.value
            }
        })
    }

    if (input.parameters) {
        let isFirst = true, value
        if (uri.search && uri.search.length > 0) {
            isFirst = false
        }
        input.parameters && input.parameters.forEach(parameter => {
            if (parameter.key) {
                value = parameter.value ? parameter.value.toString() : ''
                path += (isFirst ? '?' : '&')
                path += parameter.key + "=" + encodeURIComponent(value)
                isFirst = false
            }
        })
    }

    let data = ""
    if (input.method !== 'GET') {
        const body = input.body || {}
        const requestType = body.type || "raw"
        switch (requestType) {
            case "form-data":
                var form = new FormData()
                if (body.formData && body.formData.fields) {
                    body.formData.fields.forEach(function(item) {
                        form.append(item.field, item.value)
                    })
                }
                if (body.formData && body.formData.files) {
                    body.formData.files.forEach(function(item) {
                        form.append(item.field, item.content, item.filename)
                    })
                }
                data = form.getBuffer()
                var formHeaders = form.getHeaders()
                for (var key in formHeaders) {
                    if (formHeaders.hasOwnProperty(key)) {
                        headers[key] = formHeaders[key]
                    }
                }
                break;
            case "x-www-form-urlencoded":
                const formData = {}
                if (body.formData && body.formData.fields) {
                    body.formData.fields.forEach(item => {
                        if (item.field) {
                            formData[item.field] = item.value
                        }
                    })
                }
                data = querystring.stringify(formData)
                headers['Content-Type'] = 'application/x-www-form-urlencoded'
                break;
            //raw body
            case "raw":
                data = body.data || ""
                if (typeof data != 'string' && !Buffer.isBuffer(data)) {
                    if (body.contentType == "application/json") {
                        try {
                            data = JSON.stringify(data)
                        } catch (e) {
                            console.error(e)
                            return callback(new Error("Error parsing body.data: " + e.message))
                        }
                    } else {
                        return callback(new Error("body.data should be a string or Buffer."))
                    }
                }
                if (body.contentType) {
                    headers['Content-Type'] = body.contentType
                }
                break;
            case "file":

                break;
        }

        //set content length
        if (!headers['Content-Length']) {
            let contentLength
            if (typeof data === "string") {
                contentLength = data.length
            } else if (Buffer.isBuffer(data)) {
                contentLength = Buffer.byteLength(data)
            }
            if (typeof contentLength != "undefined") {
                headers['Content-Length'] = contentLength
            }
        }
    }

    if (input.authorization) {
        if (input.authorization.basic) {
            headers['Authorization'] = 'Basic ' + Buffer.from(input.authorization.basic.username + ':' + input.authorization.basic.password).toString('base64')
        }
    }

    const options = {
        host: uri.hostname,
        port: uri.port,
        path: path,
        method: input.method || 'GET',
        headers: headers
    }
    const inputAdvanced = input.advanced || {}
    if (inputAdvanced && inputAdvanced.encoding) {
        options.encoding = inputAdvanced.encoding
    }

    const requester = uri.protocol === 'https:' ? Https : Http
    let responseData
    const req = requester.request(options, function(resp) {
        let chunks = []
        resp.on('data', function(data) {
            chunks.push(data)
        })
        resp.on('end', function() {
            let responseBuffer = Buffer.concat(chunks)
            if (inputAdvanced.parseAsJson) {
                let responseStr
                try {
                    responseStr = responseBuffer.toString(options.encoding)
                } catch (err) {
                    err.message = `Unexpected error while converting buffer to string with encoding ${options.encoding}: ${err.message}`
                    return callback(err)
                }
                try {
                    if (responseStr) {
                        responseData = JSON.parse(responseStr)
                    } else {
                        responseData = undefined
                    }
                } catch (err) {
                    err.message = `Error while parsing response '${responseStr}' as JSON: ${err.message}`
                    return callback(err)
                }
            } else {
                if (options.encoding && options.encoding !== 'binary') {
                    responseData = responseBuffer.toString(options.encoding)
                } else {
                    responseData = responseBuffer
                }
            }
            return callback(null, {
                headers: resp.headers,
                statusCode: resp.statusCode,
                responseData: responseData
            })
        })
    })
    if (input.method !== 'GET') {
        req.write(data)
    }
    req.end()
    var errorHandled = false
    req.on('error', function(e) {
        if (!errorHandled) {
            return callback(e)
        }
    })
    var timeout = (typeof inputAdvanced.timeout == 'undefined') ? 10000 : inputAdvanced.timeout
    if (timeout > 0) {
        req.on('socket', function(socket) {
            socket.setTimeout(timeout)
            socket.on('timeout', function() {
                errorHandled = true
                req.abort()
                return callback(new Error("Response not received within specified timeout " + (timeout) + " milliseconds, aborted."))
            })
        })
    }
}

function requestHttpsAuthJSON(host, path, method, data, nonce, consumerKey, accessToken, consumerSecret, accessSecret, callback) {
    var url = {
        host: host,
        path: path,
        method: method
    }
    var timestamp = Math.floor(new Date().getTime() / 1000)
    var oauthData = {
        oauth_consumer_key: consumerKey,
        oauth_nonce: nonce,
        oauth_signature_method: "HMAC-SHA1",
        oauth_timestamp: timestamp,
        oauth_token: accessToken,
        oauth_version: "1.0"
    }
    var sigData = {}
    for (var k in oauthData) {
        sigData[k] = oauthData[k]
    }

    for (var k in data) {
        sigData[k] = data[k]
    }

    var sig = generateOAuthSignature(url, method, sigData, consumerSecret, accessSecret)
    oauthData.oauth_signature = sig
    var oauthHeader = ""
    for (var k in oauthData) {
        oauthHeader += ", " + encodeURIComponent(k) + "=\"" + encodeURIComponent(oauthData[k]) + "\""
    }
    oauthHeader = oauthHeader.substring(1)

    var req = Https.request(url, function(resp) {
        var chunks = []
        resp.on("data", function(chunk) {
            chunks.push(chunk)
        })

        resp.on("end", function() {
            var body = Buffer.concat(chunks)
            if (resp.statusCode != 200) {
                callback(null, {
                    error: resp.statusCode,
                    message: body.toString()
                })
            }
            else {
                try {
                    var parsed = JSON.parse(body.toString())
                    callback(null, parsed)
                } catch (err) {
                    callback(err)
                }
            }
        })
    })

    req.setHeader("Authorization", "OAuth" + oauthHeader)
    if (method != 'GET') {
        req.write(querystring.stringify(data))
    }
    req.end()
}

function generateOAuthSignature(url, method, data, consumerSecret, accessSecret) {
    var signingToken = encodeURIComponent(consumerSecret) + "&" + encodeURIComponent(accessSecret)

    var keys = []
    for (var d in data) {
        keys.push(d)
    }
    keys.sort()
    var output = method + "&" + encodeURIComponent('https://' + url.host + url.path) + "&"
    var params = ""
    keys.forEach(function(k) {
        params += "&" + encodeURIComponent(k) + "=" + encodeURIComponent(data[k])
    })
    params = encodeURIComponent(params.substring(1))

    return hashString(signingToken, output + params, "base64")
}

function hashString(key, str, encoding) {
    var hmac = crypto.createHmac("sha1", key)
    hmac.update(str)
    return hmac.digest(encoding)
}