'use strict'

const util = require('./util')
const _ = require('underscore')

function getError(input, message) {
    return new Error(`Assertion${(input && input.message ? " '" + input.message + "'" : "")} failed.${message ? " " + message : ""}`)
}

const getCircularReplacer = () => {
    const seen = new WeakSet()
    return (key, value) => {
        if (typeof value === "object" && value !== null) {
            if (seen.has(value)) {
                return
            }
            seen.add(value)
        }
        return value
    }
}

function format(input) {
    if (typeof input == 'object') {
        if (input == null) {
            return "null"
        }
        return "an object: " + JSON.stringify(input, getCircularReplacer())
    }
    if (typeof input == 'string') {
        return `a value "${input}" of type string`
    }
    if (typeof input == 'undefined') {
        return "undefined"
    }
    return `a value ${input} of type ${typeof input}`
}

module.exports.assertMulti = function(input, callback) {
    const inputData = input || {}
    const okAssertions = inputData.assertions
    if (okAssertions) {
        for (let i = 0; i < okAssertions.length; i++) {
            if (!okAssertions[i].condition) {
                return callback(new Error("Assertion failed: "
                    + (okAssertions[i].message || `Assertion with index #${i} evaluated to false.`)))
            }
        }
    }
    return callback()
}

module.exports.assertFalse = function(input, callback) {
    const inputData = input || {}
    if (inputData.value) {
        return callback(getError(inputData, "Expected value to be false, but got " + format(inputData.value) + "."))
    }
    return callback()
}

module.exports.assertTrue = function(input, callback) {
    const inputData = input || {}
    if (!inputData.value) {
        return callback(getError(inputData, "Expected value to be true, but got " + format(inputData.value) + "."))
    }
    return callback()
}

module.exports.assertEquals = function(input, callback) {
    const inputData = input || {}
    const ignoreArrayOrdering = inputData.ignoreArrayOrdering || false
    if (!util.isEqual(inputData.expected, inputData.actual, ignoreArrayOrdering)) {
        return callback(getError(inputData, "Expected " + format(inputData.expected) + ", but got " + format(inputData.actual) + "."))
    }
    return callback()
}

module.exports.assertNotEquals = function(input, callback) {
    const inputData = input || {}
    const ignoreArrayOrdering = inputData.ignoreArrayOrdering || false
    if (util.isEqual(inputData.expected, inputData.actual, ignoreArrayOrdering)) {
        return callback(getError(inputData, "Expected " + format(inputData.expected) + " to be not equal to " + format(inputData.actual) + "."))
    }
    return callback()
}

module.exports.assertEmpty = function(input, callback) {
    const inputData = input || {}
    if (util.isSet(inputData.value)) {
        return callback(getError(inputData, "Expected value to be empty, but found to be " + format(inputData.value) + "."))
    }
    return callback()
}

module.exports.assertNotEmpty = function(input, callback) {
    const inputData = input || {}
    if (!util.isSet(inputData.value)) {
        return callback(getError(inputData, "Expected value to be not empty, but found to be empty (" + format(inputData.value) + ")."))
    }
    return callback()
}

module.exports.fail = function(input, callback) {
    const inputData = input || {}
    return callback(getError(inputData))
}
