const path = require('path')
const _ = require('underscore')
const async = require('async')

//engine
const subflow = require('../core/subflow')
const loader = require(path.join(ENGINE_DIR, 'loader'))
//util
const util = require('./util')

//interface id
const INTERFACE_ID = "core/testing/testinterface"
//default global timeout
const DEFAULT_TIMEOUT = 30000
//default scan depth
const DEFAULT_SCAN_DEPTH = 0
//global object to associate TestRunner instance with flow/step
const runner = {}

/**
 * Module entry point.
 */
module.exports.run = function(input, callback, options) {
  input = input || {}
  const stepId = options.filename + ':' + options.step
  input.options = input.options || {}

  if (typeof input.testsFolder == "undefined") {
    return callback(new Error("testsFolder should be set"))
  }
  //serve
  if (!runner[stepId]) {
    runner[stepId] = new TestRunner()
  }
  return runner[stepId].run(input, options, callback)
}

function TestRunner() {
  this.initialized = false
  this.handlerFolder = ''
  this.testFlows = []
}

function formatTests(tests) {
  return tests.map(function(test) {
    return { flow: test.flow, error: test.error.message }
  })
}

TestRunner.prototype.run = function(input, options, callback) {
  const self = this
  this.init(input, function(err) {
    if (err) {
      return callback(err)
    }
    if (!self.testFlows.length) {
      return callback(null, {
        success: true,
        totalTests: 0,
        totalFailed: 0,
        report: {
          all: [],
          failed: []
        }
      })
    }
    const failedTests = []
    const totalTests = []
    const testsIdx = {}
    //run tests in parallel or series?
    let iterFn = async.eachSeries
    if (input.options.parallel) {
      if (input.options.limit) {
        iterFn = function(arr, iterator, cb) {
          async.eachLimit(arr, input.options.limit, iterator, cb)
        }
      } else {
        iterFn = async.each
      }
    }
    iterFn(self.testFlows, function(testFlow, cb) {
      console.log("Executing test:", testFlow.flow)
      const testId = Math.random() * 10000000000000000
      testsIdx[testId] = {
        completed: false,
        timedout: false
      }
      const flowName = testFlow.flow
      const timeout = typeof input.options.timeout == 'undefined' ? DEFAULT_TIMEOUT : input.options.timeout
      const tidx = setTimeout(function() {
        if (!testsIdx[testId].completed) {
          testsIdx[testId].timedout = true
          failedTests.push({
            flow: flowName,
            error: new Error("Error: Timeout of " + timeout + "ms exceeded.")
          })
          return handleCb()
        }
      }, timeout)

      function handleCb() {
        if (testsIdx[testId].completed) {
          return
        }
        clearTimeout(tidx)
        testsIdx[testId].completed = true
        totalTests.push(flowName)
        if (totalTests.length >= self.testFlows.length) {
          const failedTestReport = formatTests(failedTests)
          console.log("Total tests: ", totalTests.length)
          if (failedTests.length) {
            console.error(`Failed ${failedTests.length} out of ${totalTests.length}`)
            console.error("Failed tests: ", failedTestReport)
            process.exitCode = 1
          } else {
            console.log("All pass!")
          }
          return callback(null, {
            success: failedTests.length == 0,
            totalTests: totalTests.length,
            totalFailed: failedTestReport.length,
            report: {
              all: totalTests,
              failed: failedTestReport
            }
          })
        }
        cb()
      }

      self.executeTestFlow(flowName, {}, function(err, result) {
        if (err) {
          failedTests.push({
            flow: flowName,
            error: err
          })
        }
        return handleCb()
      }, options)
    })
  })
}

TestRunner.prototype.init = function(input, callback) {
  const self = this
  if (this.initialized) {
    return callback()
  }
  const testsFolder = input.testsFolder ? util.normalizePrefix(input.testsFolder) : ''
  this.testsFolder = ROOT_DIR + '/' + testsFolder
  const config = input.options
  const scanDepth = typeof config.scanDepth == 'number' ? Math.abs(config.scanDepth) : DEFAULT_SCAN_DEPTH
  this.loadTests(scanDepth, function(err, response) {
    if (err) {
      return callback(err)
    }
    self.testFlows = response.testFlows
    self.testFlows.sort(function(a, b) { return b.filePath.localeCompare(a.filePath) })
    self.initialized = true
    return callback()
  })
}

/**
* Scan directory and load test flows
*
* @param {*} level
* @param {*} callback
*/
TestRunner.prototype.loadTests = function loadTests(level, callback) {
  console.log("Scanning for tests..")
  const testFlows = []
  util.listFileTree(this.testsFolder, function(err, fileTree) {
    if (err) {
      return callback(err)
    }
    util.iterateAsyncParallel(fileTree, function(filePath, next) {
      const parts = path.parse(filePath)
      if (parts.ext !== '.flw') {
        return next()
      }
      const fileName = filePath.slice(ROOT_DIR.length + 1)
      loader.loadFlow({ file: filePath }, function(err, flow) {
        if (err) {
          console.error("Error loading flow file: " + fileName, err)
          return next()
        }
        const meta = flow.meta || {}
        if (meta.inherits !== INTERFACE_ID) {
          console.error("The flow file " + fileName + " is not inheriting " + INTERFACE_ID + ", ignoring.")
          return next()
        }
        const annotation = meta.annotation || {}
        if (annotation.disabled === "true") {
          console.error("The test handler " + fileName + " is marked as disabled in the config, ignoring.")
          return next()
        }
        const handler = {
          flow: fileName,
          filePath: filePath,
          tags: annotation.tags || [],
          inputSchema: flow.startStep.inputSchema,
          outputSchema: flow.endStep.outputSchema
        }
        testFlows.push(handler)
        return next()
      })
    }, function(err) {
      return callback(err, {
        testFlows: testFlows,
      })
    })
  }, { level: level })
}

TestRunner.prototype.executeTestFlow = function(flowFile, inputParameters, callback, options) {
  subflow.execute({
    flow: {
      file: flowFile,
      input: inputParameters
    }
  }, callback, options)
}