var fs = require('fs')
var path = require('path')
var isAbsolute = require('path-is-absolute')
var _ = require('underscore')

/**
 * Utility method to list files under a directory.
 *
 * @param {String} directory
 * @param {Function} callback
 * @param {Object} [options]
 * @param {Number} [level]
 */
function listFileTree(directory, callback, options, level) {
    var files = [] //array of files
    var tree = {} //hashmap of files
    level = level || 0
    options = options || {}
    //read the directory files
    fs.readdir(directory, function(err, fileList) {
        if (err) {
            return callback(err)
        }
        iterateAsyncParallel(fileList, function(file, next) {
            var filePath = path.join(directory, file)
            fs.stat(filePath, function(err, stat) {
                if (err) {
                    return next(err)
                }
                if (stat && stat.isDirectory()) {
                    //if a level is specified in the options, dig only that deeper
                    if (typeof options.level !== 'undefined' && level >= options.level) {
                        return next()
                    }
                    tree[file] = []
                    //recursively get the files inside the directory
                    listFileTree(filePath, function(err, dirFiles, dirTree) {
                        if (err) {
                            return next(err)
                        }
                        files = files.concat(dirFiles)
                        tree[file] = dirTree
                        //go to next file
                        return next()
                    }, options, level + 1)
                } else {
                    //add the file to the file path
                    files.push(filePath)
                    tree[file] = filePath
                    return next()
                }
            })
        }, function() {
            return callback(null, files, tree)
        })
    })
}

/**
 * Utility method to iterate through a collection asynchronously.
 *
 * @param {*} collection
 * @param {*} fn
 * @param {*} callback
 */
function iterateAsyncParallel(collection, fn, callback) {
    if (!collection) {
        return callback()
    }
    var keys = Object.keys(collection)
    var count = 0, len = keys.length
    if (len === 0) {
        return callback()
    }
    for (var i = 0; i < len; i++) {
        fn(collection[keys[i]], function(err) {
            if (err) {
                return callback(err)
            }
            if (++count === len) {
                return callback()
            }
        })
    }
}

/**
 * add a / to the beginning and remove trailing /
 */
function normalizePrefix(prefix) {
    if (prefix[0] !== '/') {
        prefix = "/" + prefix
    }
    //and remove a trailing /
    if (prefix.lastIndexOf('/') === prefix.length - 1) {
        prefix = prefix.slice(-1)
    }
    return prefix
}

//@TODO - copied from expression helper. Generalize it.
function isSet(obj) {
    var isSet = true
    switch (typeof obj) {
        case 'string':
            isSet = obj.length > 0
            break
        case 'boolean':
            isSet = obj
            break
        case 'number':
            isSet = obj > 0
            break
        case 'object':
            if (obj === null) {
                isSet = false
            }
            else if (+obj.length === obj.length) { //array
                isSet = (obj.length > 0)
            } else { //obj
                isSet = Object.keys(obj).length > 0
            }
            break
        case 'undefined':
            isSet = false
            break
        case 'function': // unusual?
            isSet = true
            break
        default:
            isSet = false
    }
    return isSet
}

function isEqual(obj, obj2, ignoreArrayOrdering) {
    if (ignoreArrayOrdering) {
        if (_.isArray(obj) && _.isArray(obj2)) {
            return (obj.length === obj2.length)
                && (obj.length === _.intersection(obj, obj2).length)
        }
    }
    return _.isEqual(obj, obj2)
}

module.exports = {
    iterateAsyncParallel: iterateAsyncParallel,
    listFileTree: listFileTree,
    normalizePrefix: normalizePrefix,
    isSet: isSet,
    isEqual: isEqual,
}
