/**
 * Author Murukesh Sadasivan (murukesh@codeflow.co)
 *
 * Copyright 2016 Codeflow Private Limited
 *
 * MIT Licensed
 */
'use strict'
/**
 * Generator of Javascript code for JIT compilation of flows. A flow gets converted to a class
 * which is instantiated for every execution of the flow.
 *
 * Due to usage of template strings, this code requires node.js >= @4.0 to run.
 */
const Path = require('path')
const async = require('async')
const _ = require('underscore')
const Flow = require('../flow')
const _Error = require('../error')
const Constants = require('../constants')
const helper = require('./helper')

const header =
    `/**
* Auto-generated by Codeflow Engine JIT compiler.
*/
// 'use strict'
//Common handle used for error handling
var _inputError = null
function _throwError(msg) {
   _inputError = new Error(msg)
}
`

function JsCompiler(flow, params) {
    params = params || {}
    this.flow = flow
    this.className = params.className || 'Flow'
    this.isDebug = params.isDebug
    this.params = params
    this.steps = {}
    this.codeHeader = ''
    this.codeBody = ''
    this.isTrigger = false
}

JsCompiler.prototype.generate = function(cb) {
    const startStep = _.find(this.flow.steps, step => {
        return step.isStart || step.isTrigger
    })
    this.isTrigger = startStep.isTrigger
    //generate the class/frame to represent the flow
    this.generateFlowClass()
    //generate methods for each individual step in the flow
    this.generateStep(startStep, err => {
        //compile the generated code inside a closure
        const script = header + this.codeHeader + '\n' + this.codeBody
        try {
            //create the wrapper function around the generated script
            const closureFn = new Function("helper", "__methods", "_Error", "flow", script + '\nreturn Flow')
            //now call the wrapper with arguments to get the Flow class
            const compiled = closureFn.apply(null, [helper, helper._getMethodContext(), _Error, this.flow])
            //callback
            return cb(null, compiled)
        } catch (e) {
            console.error("Error while compiling flow: ", e)
            // console.log("script", script)
            return cb(e)
        }
    })
}

JsCompiler.prototype.generateFlowClass = function() {
    const className = ''
    const steps = this.flow.steps
    let stepsInit = ''
    _.each(steps, step => {
        const incoming = step.incoming
        if (incoming.length > 1 || step.meta.autoLog) {
            if (incoming.length > 1) {
                stepsInit += `\nthis.steps["${step.id}"] = {invokeCount: ${incoming.length}, activePathCount: ${incoming.length}, autoLog: ${step.meta.autoLog}}`
            } else {
                stepsInit += `\nthis.steps["${step.id}"] = {autoLog: ${step.meta.autoLog}}`
            }
        }
    })
    //@TODO - revisit global handlers (inputError & error)
    let debugProtoCode = '' //code to insert as prototype level functions if started in debug mode
    let debugConstructorCode = '' //code to be inserted in constructor if started in debug mode
    if (this.isDebug) {
        debugConstructorCode =
            `
        this.id = helper.getId(parent)
        this.stepMap = {}
        this.parent = parent
        this.parentStep = parentStep
        var parentActMap = parent && parent.stepMap ? parent.stepMap[parentStep] : {}
        if (parentActMap.mode == 1) {
            this.mode = 1 //STEP
        }
        this.rootId = parent? parent.rootId: this.id
        this.subInstances = {}
        `
        debugProtoCode =
            `
         ${this.className}.prototype.addSubInstance = function(instance) {
             this.subInstances[instance.id] = instance
         }

         ${this.className}.prototype.removeSubInstance = function(instance) {
            delete this.subInstances[instance.id]
         }
         `
    }

    let groupLookup = ''
    if (this.params.scope) {
        this.params.scope.split(Constants.GROUP_ID_SEPARATOR).forEach((id) => {
            groupLookup += '.steps.' + id + '.data'
        })
    }

    this.codeHeader = `
        /* ${this.flow.fileName} */
        function ${this.className}(engine, params, parent, parentStep) {
            this.engine = engine
            this.flow = flow${groupLookup || ''}
            this.params = params || {}
            this.isTrigger = ${this.isTrigger}
            this.tick = 0 //used
            this.steps = {
                global: global
            }
            if (parent) {
                this.parent = parent
                if (parent.steps) {
                    var keys = Object.keys(parent.steps)
                    for (var i = 0; i < keys.length; i++) {
                        this.steps[keys[i]] = parent.steps[keys[i]]
                    }
                }
            }
            if (params.data) {
                this.steps[params.id] = {
                    scope: params.data
                }
            }
            this.vars = parent? parent.vars: (params.vars ? params.vars : {})
            this.cbCalled = false
            ${stepsInit}
            ${debugConstructorCode}
        }
        ${debugProtoCode}

        ${this.className}.prototype.handleOutput = function(stepId, error, output, vars) {
            if (typeof this.steps[stepId] == 'undefined') {
                this.steps[stepId] = {}
            }
            var step = this.steps[stepId]
            if (error) {
                var stepPath =  stepId + ' (' + this.flow.fileName + ')'
                this.steps[stepId].error = helper.wrapError(error, stepPath)
            }
            this.steps[stepId].output = output
            if (vars) {
                var keys = Object.keys(vars)
                for (var i = 0; i < keys.length; i++) {
                    this.vars[keys[i]] = vars[keys[i]]
                }
            }
        }
        `
    if (!this.isTrigger) {
        this.codeBody +=
            `
            ${this.className}.prototype.doCallback = function(error, output) {
                if (this.cbCalled) {
                    console.error("Callback already invoked. Ignoring.")
                    if (error) {
                        console.error(error)
                    }
                    return
                }
                if (error) {
                    // console.error(error)
                }
                this.cbCalled = true
                return this.params.callback && this.params.callback(error, output)
            }
            `
    } else {
        this.codeBody +=
            `
            ${this.className}.prototype.doCallback = function(error, output) {
                if (error) {
                    console.error(error)
                }
                //return this.params.callback && this.params.callback(error, output)
            }`
    }
}

JsCompiler.prototype.generateGroupClass = function(step, params, callback) {
    const data = step.data
    const id = helper.makeScopeId(step.id, this.params.scope)
    const fileName = this.params.scope ? this.params.file : (this.params.name || this.params.file)
    const uniqFilename = fileName.slice(-4) === ".grp" ? fileName : fileName + '.grp'
    const groupFileName = id + "@" + uniqFilename
    new Flow().load(groupFileName, data, (err, flow) => {
        if (err) {
            return callback(err)
        }
        const groupCompiler = new JsCompiler(flow, {
            className: helper.makeClassId(id),
            scope: id,
            file: uniqFilename,
            name: groupFileName
        })
        const startStep = _.find(flow.steps, step => {
            return step.isStart || step.isTrigger
        })
        groupCompiler.generateFlowClass()
        groupCompiler.generateStep(startStep, (err) => {
            if (err) {
                return callback(err)
            }
            return callback(null, groupCompiler.codeHeader + groupCompiler.codeBody)
        })
    })
}

/**
 * Generate code for the given step. Recursively calls itself for generating subsequent steps (from outgoing transitions).
 */
JsCompiler.prototype.generateStep = function(step, callback) {
    step.load((err, stepDefinition) => {
        this.generateFunctionDefinition(step, err, (err, data) => {
            if (data && data.body) {
                this.codeBody += data.body
            }
            if (data && data.header) {
                this.codeHeader += data.header
            }
            if (step.outgoing.length > 0) {
                async.eachSeries(step.outgoing, (outgoing, next) => {
                    this.generateStep(outgoing.to, next)
                }, () => {
                    callback(null)
                })
            } else {
                callback(null)
            }
        })
    })
}

/**
 * Generate a javascript function definition that represent a step and its outgoing transitions.
 */
JsCompiler.prototype.generateFunctionDefinition = function(step, err, callback) {
    if (this.steps[step.id]) {
        return callback(null)
    }
    this.steps[step.id] = true
    //name for function
    const fnName = step.isStart || step.isTrigger ? 'start' : `step_${step.id}`
    const fnInputName = step.isStart || step.isTrigger ? '_input_start' : `_input_${step.id}`
    //generating code depends on the outgoing transitions to decide which step(s) to be invoked next in the chain.
    const outgoing = step.outgoing
    //compile jst input
    const inputCompiled = helper.compileJst(step.inputBinding)
    const source = step.sourcePath ? Path.join(step.sourcePath, step.source) : ''
    const meta = step.meta
    //arrays to hold code for various invocations
    const calls = []
    const errorCalls = []
    const alwaysCalls = []
    const otherwise = []
    //string to hold the generated function
    let funcStr = ''
    let headerStr = ''

    //flag to mark if otherwise transitions should be triggered
    let preventOtherwise = false
    //count of conditional transitions. to keep track of number of conditional transitions that are triggered.
    //if none, then otherwise transitions, if any might be called.
    let conditionsCount = 0
    //first pass, compute flags
    outgoing.forEach(transition => {
        if (transition.type === 'condition') {
            conditionsCount++
        } else if (!transition.type) {
            //if there is at least one regular transition, otherwise transitions are not triggered
            preventOtherwise = true
        }
    })
    //otherwise transitions are considered only if there is at least one conditional transition
    const allowOtherwise = (conditionsCount > 0 || outgoing.length === 1) && !preventOtherwise

    // junction calls or exclusion calls - what happens when a transition is not triggered? 
    // it still needs to make some calls so as to disable subsequent paths as well as free any join steps
    // from waiting for it.
    const nonErrorJunctionCalls = []
    const errorJunctionCalls = []

    //second pass - generate call for every transitions except the conditional ones.
    outgoing.forEach(transition => {
        let callFn = ''
        if (transition.type === 'condition') {
            //dont do anything, will be handled in second pass.
        } else if (transition.type === 'otherwise') {
            otherwise.push(transition)
        } else if (transition.type === 'error') {
            callFn = `self.step_${transition.to.id}()`
            errorCalls.push(callFn)
        } else if (transition.type === 'always') {
            callFn = `self.step_${transition.to.id}()`
            alwaysCalls.push(callFn)
        } else {
            callFn = `self.step_${transition.to.id}()`
            calls.push(callFn)
        }
        if (transition.type !== 'error' && transition.type !== 'always') {
            //nonErrorJunctionCalls - calls to execute (to disable steps that are not in the error path) when there is an exception
            //transitions that are not error or always will never gets triggered
            nonErrorJunctionCalls.push(this.getJunctionCalls(transition))
        } else {
            errorJunctionCalls.push(this.getJunctionCalls(transition))
        }
    })
    //generate string from array of calls
    let nonErrorJunctionCallsStr = nonErrorJunctionCalls.join('\n')
    let errorCallsStr = errorCalls.join('\n')
    let alwaysCallsStr = alwaysCalls.join('\n')

    //if there is either an error path or always path
    if (errorCallsStr || alwaysCallsStr) {
        errorCallsStr += ('\n' + nonErrorJunctionCallsStr)
    }

    //third pass - conditional transitions are computed after as the error handling code is
    //required to be computed before as there could be errors while evaluating conditional expressions
    let errorHandling = ''
    if (errorCalls.length || alwaysCalls.length) {
        errorHandling = `
        self.handleOutput("${step.id}", _inputError)
        ${errorCallsStr}
        ${alwaysCallsStr}
        return
        `
    } else {
        errorHandling = `
        self.doCallback(helper.wrapError(_inputError, "${step.path}"))
        return`
    }
    let trCnt = 0
    //third pass - generate code for transitions with conditions
    outgoing.forEach(transition => {
        if (transition.type === 'condition') {
            trCnt++
            const condition = transition.condition
            let expr
            //handle empty conditions
            if (typeof condition == 'undefined' || (typeof condition == 'string' && condition.length === 0)) {
                expr = 'true'
            } else {
                expr = helper.compileExpression(transition.condition)
            }
            let elseFn = ''
            //decrement conditionsCount if the transition doesn't happen
            //this is to trigger otherwise if conditionsCount reaches 0
            if (allowOtherwise) {
                elseFn += '_conditionsCount--\n'
            }
            elseFn += this.getJunctionCalls(transition, true)
            const trid = getTrId(transition)
            let callFn = `
            _inputError = null
            var transition_${trCnt} = ${expr}
            if (_inputError) {
                _inputError.message = 'Error occurred while evaluating condition at transition from [${transition.from.id}] to [${transition.to.path}]. Error: [' + _inputError.message + ']'
                ${errorHandling}
            } else {
                if (transition_${trCnt}) {
                    self.step_${transition.to.id}()
                } else {
                    ${elseFn}
                }
            }`
            //added to the beginning
            calls.unshift(callFn)
        }
    })

    //otherwise transitions
    let callsStr = allowOtherwise ? `var _conditionsCount = ${conditionsCount}` : ''
    const otherwiseCalls = []
    if (allowOtherwise) {
        otherwise.forEach(transition => {
            otherwiseCalls.push(`
            if (_conditionsCount === 0) {
                self.step_${transition.to.id}()
            } else {
                ${this.getJunctionCalls(transition, true)}
            }`)
        })
    } else if (otherwise.length) {
        //in case of otherwise transitions existing, but are not triggered due to conditional transitions 
        //happening, generate exclusion calls (junction calls)
        otherwise.forEach(transition => {
            otherwiseCalls.push(this.getJunctionCalls(transition, true))
        })
    }
    //join calls array to string
    callsStr += ('\n' + calls.join('\n'))
    //and append otherwise calls if any
    if (otherwiseCalls.length > 0) {
        callsStr += otherwiseCalls.join('\n')
    }
    //finally add calls to handle transitions that were not triggered
    //when there are no errors
    if (errorCalls.length) {
        callsStr += errorJunctionCalls.join('\n')
    }
    //now start generating the actual function code.
    //but for steps that have multiple incoming paths, it's not immediately executed or may be never executed at all.
    //for steps with more than one incoming paths, a join condition is triggered which makes sure that all the paths
    //leading to that step are either evaluated or is unreachable (due to that path not being taken due to error/condition etc).
    let joinCode = ''
    const joinCount = step.incoming.length
    if (joinCount > 1) {
        //generate exclusion calls in case in case the step is not invoked.
        //this happens if all the paths leading to the step becomes inaccessible.
        const postJoinCalls = []
        step.outgoing.forEach(transition => {
            const tmp = this.getJunctionCalls(transition, true)
            if (tmp) {
                postJoinCalls.push(tmp)
            }
        })
        //what happens after join in case the step is not invoked.
        let postJoinDiscardCode = ''
        if (postJoinCalls.length > 0) {
            postJoinDiscardCode =
                `if (step.invokeCount === 0) {
                ${postJoinCalls.join('\n')}
            }`
        }
        joinCode = `
        var step = this.steps["${step.id}"]
        step.invokeCount--;
        var continueStep = step.invokeCount === 0 && step.activePathCount > 0
        if (!continueStep) {
            ${postJoinDiscardCode}
            return
        }`
    }
    const logInput = meta && meta.autoLog ? `helper.logInput("${step.path}", input)` : ''
    const logOutput = meta && meta.autoLog ? `helper.logOutput("${step.path}", self.steps["${step.id}"].output)` : ''

    //If it's an end step, handle separately (simpler way)
    if (step.isEnd) {
        let prefix = ''
        if (joinCode) {
            prefix = `
            if (typeof this.steps["${step.id}"] === undefined) {
                this.steps["${step.id}"] = {}
            }`
        }
        funcStr += `
        ${this.className}.prototype.${fnName} = function() {
            ${prefix}${joinCode}
            var data = this.steps
            _inputError = null
            var input = ${inputCompiled}
            ${logInput}
            if (_inputError) {
                _inputError = helper.makeError(_inputError, "${step.path}")
            }
            this.doCallback(_inputError, input)
        }\n`
        return callback(null, { body: funcStr })
    }

    //all other steps get handled below
    let callsStmt = ''
    let defMissingError = ''
    let errorOnInput = ''

    //if there is no error or always transition
    if (!alwaysCalls.length && !errorCalls.length) {
        if (err) {
            defMissingError = `return self.doCallback(helper.wrapError(\`${err.message}\`, "${step.path}"))`
        }
        errorOnInput = `
            if (_inputError) {
                return self.doCallback(helper.wrapError(_inputError, "${step.path}"))
            }
        `
        callsStmt = `
            if (_error) {
                return self.doCallback(helper.wrapError(_error, "${step.path}"))
            } else {
                ${callsStr}
            }
        `
    } else {
        if (err) {
            defMissingError = `
            self.handleOutput("${step.id}", \`${err.message}\`)
            ${errorCallsStr}
            ${alwaysCallsStr}
            return
            `
        }
        errorOnInput = `
            if (_inputError) {
                self.handleOutput("${step.id}", _inputError)
                ${errorCallsStr}
                ${alwaysCallsStr}
                return
            }
        `
        callsStmt = `
            if (_error) {
                ${errorCallsStr}
                ${alwaysCallsStr}
            } else {
                ${callsStr}
                ${alwaysCallsStr}
            }
        `
    }

    let triggerStmt = ''
    if (step.isTrigger) {
        triggerStmt =
            `var triggerInstance = helper.cloneInstance(${this.className}, thisInstance, '${step.id}');
         triggerInstance.vars = vars || {}
         var self = triggerInstance
         var data = self.steps
         `
    }

    funcStr += `
    ${this.className}.prototype.${fnName} = function() {
            var self = this
            ${joinCode}
            ${defMissingError}
            var data = this.steps
            _inputError = null
            var input = ${inputCompiled}
            ${logInput}
            self.steps["${step.id}"] = {}
            ${errorOnInput}
        `
    this.getStaticSnippet(step, inputCompiled, callsStmt, (err, snippet) => {
        if (snippet) {
            if (snippet.async) {
                funcStr +=
                    `
                ${snippet.async}
                `
            } else {
                funcStr +=
                    `var _error = null
                ${snippet.sync}
                ${logOutput}
                ${callsStmt}
                `
            }
        } else {
            if (step.isTrigger) {
                funcStr += '\nvar thisInstance = self\n'
            }
            funcStr += `var Cls = null\n`
            if (step.isGroup) {
                funcStr += `Cls = ${helper.makeClassId(step.id, this.params.scope)} \n`
            }
            let scope = this.params.scope ? `'${this.params.scope}'` : 'null'
            if (step.flowfile) {
                funcStr +=
                    `helper.callFlow(this, {id: '${step.id}', packageName: '${step.packageName || ""}', flowfile: '${source}', input: input, Cls: Cls, scope: ${scope}}, function(_error, output, vars) {
                    ${triggerStmt}
                    self.handleOutput("${step.id}", _error, output, ${!step.isTrigger ? 'vars' : 'null'})
                    ${logOutput}
                    ${callsStmt}
                })`
            } else {
                funcStr +=
                    `helper.callJs(this, {id: '${step.id}', modulePath: '${source}', method: '${step.method}', input: input, Cls: Cls, scope: ${scope}}, function(_error, output, vars) {
                    ${triggerStmt}
                    self.handleOutput("${step.id}", _error, output, ${!step.isTrigger ? 'vars' : 'null'})
                    ${logOutput}
                    ${callsStmt}
                })`
            }
        }
        funcStr += '}\n'
        if (step.isGroup) {
            this.generateGroupClass(step, {}, (err, groupScript) => {
                headerStr += groupScript
                return callback(null, { body: funcStr, header: headerStr })
            })
        } else {
            return callback(null, { body: funcStr, header: headerStr })
        }
    })
}

/**
 * Get a readable id for transition
 */
const getTrId = function(transition) {
    return transition.from.id + ':' + transition.to.id
}

/**
 * When a transition does not trigger, the outgoing steps and their whole chain of
 * paths are disabled until we reach a junction - i.e. where the step has an
 * incoming path not originating from the the disabled path.
 *
 * There can be one or more of such junction steps and below method finds and
 * generate calls for them.
 */
JsCompiler.prototype.getJunctionCalls = function(transition, force) {
    let fn = ''
    if (force || (transition.from.outgoing.length > 1 || transition.to.incoming.length > 1)) {
        const junctionSteps = this.findJunctionSteps(transition)
        if (junctionSteps) {
            junctionSteps.forEach(junctionStep => {
                let step = junctionStep.step
                for (var i = 0; i < junctionStep.count; i++) {
                    fn +=
                        `\nself.steps["${step.id}"].activePathCount--;
                   self.step_${step.id}()`
                }
            })
        }
    }
    return fn
}

/**
 * Search to find junction steps - ones which have an incoming path not originating from the given transition.
 */
JsCompiler.prototype.findJunctionSteps = function(transition) {
    const steps = [] //to hold list of junction steps
    const staleList = {} //list of paths whose ultimate origin is from the given transition
    const processed = {} //to store processed steps
    const q = [] //temporary queue
    q.push(transition)
    //init
    staleList[getTrId(transition)] = true
    //process paths beginning with the first one
    while (q.length) {
        const item = q.shift()
        //find the outgoing step
        const step = item.to
        if (processed[step.id]) {
            processed[step.id]++ //keep track of number of times the step was hit
            continue //dont process further if it's already processed
        } else {
            processed[step.id] = 1 //init number of times the step was hit (1 as of now)
        }
        let isStale = false //flag to mark if the step is stale
        if (step.incoming.length > 1) {
            //if there are more than one incoming path, iterate and check if all of them are stale
            const activeList = _.filter(step.incoming, transition => {
                return !staleList[getTrId(transition)]
            })
            //if not, then add it as a junction step
            if (activeList && activeList.length) {
                steps.push(step)
            } else {
                isStale = true
            }
        } else {
            isStale = true
        }
        if (isStale) {
            //if there is only one incoming path, step is also stale
            //and add all the outgoing paths as stale as well and continue processing the chain
            step.outgoing.forEach(transition => {
                staleList[getTrId(transition)] = true
                q.push(transition)
            })
        }
    }
    //return the junction steps and the number of times it was hit
    return _.map(steps, step => {
        return {
            step: step,
            count: processed[step.id]
        }
    })
}

/**
 * To improve performance (avoiding multiple function calls), few commonly used steps are generated as static snippets.
 * They are stored in js/snippets.js file and are returned if specified.
 */
JsCompiler.prototype.getStaticSnippet = function(step, inputCompiled, callsStmt, callback) {
    helper.getSnippet(step, inputCompiled, callsStmt, this.params, 'sync', (error, result) => {
        if (result || error) {
            return callback(error, result)
        }
        helper.getSnippet(step, inputCompiled, callsStmt, this.params, 'async', (error, result) => {
            return callback(error, result)
        })
    })
}

module.exports = JsCompiler